# $Id$

# /** @file Bentokit Retromod scheduled media and spot update to Bentokit Flywire playout client. */

use playout::Client;

# /** Update module for the scheduled sponsorship data that goes to and from Flywire on the client machine. */
package playout::Client::Flywire;

## Specify parent for inheritance.
@ISA = (playout::Client);

use strict;
use playout::Storage;
use File::Path;

our $AUTOLOAD;

## Permitted properties.
my %fields = (
	ip => undef,
	log_path => undef,
	media_path => undef,
	schedule_path => undef,
	username => undef,
	onair => undef,
	storage_path => undef,
	cache_path => undef,
	storage_method => undef,
	dbh => undef,
	store => undef,
	trafficStore => undef,
	productionStore => undef,
	flywireCache => "flywire",
	flywireAutoMediaCache => "AutoMedia",
	flywireAutoMediaMusicCache => "AutoMedia/Music",
	flywireLogCache => "Log",
	flywireScheduleCache => "Schedule",
	flywireScheduledMediaCache => "ScheduledMedia",
	flywireShowMediaCache => "ShowMedia",
	flywireStationMediaCache => "StationMedia",
);

# /** Destructor. */
sub DESTROY {
	#Retromod::DB->disconnect($self->dbh);
	my ($self) = @_;
}

# /** Constructor. */
sub new {
	# Use our inherited AUTOLOAD but allow set and get of our params.
	my $class = shift;
	my $self = $class->SUPER::new();
	my($element);
	foreach $element (keys %fields) {
		$self->{_permitted}->{$element} = $fields{$element};
	}
	@{$self}{keys %fields} = values %fields;

	# Load recurring objects.
	#$self->store = playout::Storage->new();
	#$self->trafficStore = traffic::Storage->new();
	#$self->dbh = Retromod::DB->connect;
	return $self;
}

# /** Help page. */
sub help {
	my($self, $args) = @_;

	my $help_message = <<EOH;
This is the help for the Flywire client module for the playout component of Retromod.

Usage: update.pl playout method Flywire [ arguments ]

  Method:
  help
    Returns this help page.
  process_all
    Updates traffic, ... on the remote Flywire client.
  process_all_traffic
    Updates logs, schedule and scheduled media on the remote Flywire client.
  process_logs
    Updates logs on the remote Flywire client.
  process_scheduled_media
    Updates scheduled media on the remote Flywire client.
  process_schedule
    Updates the schedule on the remote Flywire client.

EOH
    print "$help_message\n";
}

# /** Update remote Flywire client. */
sub process_all {
	my $self = shift;

	$self->process_all_traffic();
}

# /** Update traffic processes on remote Flywire client. */
sub process_all_traffic {
	my $self = shift;

	$self->check_component_exists('traffic');

	$self->process_logs();
	$self->process_scheduled_media();
	$self->process_schedule();
}

# /** Update logs for remote Flywire client. */
sub process_logs {
	my $self = shift;
	$self->check_component_exists('traffic');

	# Ingest logs only if studio was designated onair.
	if ($self->onair) {
		print "Remote client '".$self->ip."' was designated as on-air. Proceeding...\n";
		$self->sync_remote_logs();
		$self->ingest_logs();
	} else {
		print "Remote client '".$self->ip."' was not designated as on-air. Logs not ingested.\n";
	}
}

# /** Update scheduled media for remote Flywire client. */
sub process_scheduled_media {
	my $self = shift;
	$self->check_component_exists('traffic');
	$self->cache_scheduled_media();
	$self->sync_remote_scheduled_media();
}

# /** Update schedule for remote Flywire client. */
sub process_schedule {
	my $self = shift;
	$self->check_component_exists('traffic');
	$self->cache_schedule();
	$self->sync_remote_schedule();
}

# /** Update local cache for scheduled media. */
sub cache_scheduled_media {
	my $self = shift;
	$self->purge_cache_scheduled_media();
	$self->update_cache_scheduled_media();
}

sub purge_cache_scheduled_media {
	my $self = shift;

	# Purge previously cached media on the server.

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireScheduledMediaCache;
	$path = File::Spec->canonpath($path);

    print "Purging local scheduled media cache..."."\n";
    my $command = 'rm -R -f '.Retromod::Global->escape_string($path).'/*';
    print $command."\n";
    system($command);
    print "Purged local scheduled media cache."."\n";
}

sub update_cache_scheduled_media {
	my $self = shift;

	if (!defined $self->trafficStore) {
		$self->open_remote_storage('traffic');
	}

	# Extract scheduled media from data storage.
    print "Updating local scheduled media cache..."."\n";
	my $query = <<EOQ;
		SELECT DISTINCT A.$self->{trafficStore}->{field_campaignaudio_name} AS name,
			A.$self->{trafficStore}->{field_campaignaudio_mediaID} AS mediaID,
			A.$self->{trafficStore}->{field_campaignaudio_mimetype} AS mimetype
		FROM (`$self->{trafficStore}->{table_traffic_spot}` S,
			`$self->{trafficStore}->{table_traffic_campaignaudio}` A,
			`$self->{trafficStore}->{table_traffic_contract}` O
		)
		WHERE (S.$self->{trafficStore}->{field_spot_dateTime} >= TIMESTAMP(NOW() - INTERVAL 2 HOUR)
			&& S.$self->{trafficStore}->{field_spot_dateTime} < TIMESTAMP(NOW() + INTERVAL 4 DAY)
		)
		AND S.$self->{trafficStore}->{field_spot_playedDateTime} IS NULL
		AND (O.$self->{trafficStore}->{field_contract_contractStatus} = 'Approved'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Invoiced'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Paid'
		)
		AND A.$self->{trafficStore}->{field_campaignaudio_productionStatus} = 'Approved'
		AND S.$self->{trafficStore}->{field_spot_inactive} != 'true'
		AND A.$self->{trafficStore}->{field_campaignaudio_filename} != ''
		AND A.$self->{trafficStore}->{field_campaignaudio_mediaID} = S.$self->{trafficStore}->{field_spot_mediaID}
		AND A.$self->{trafficStore}->{field_campaignaudio_contractID} = O.$self->{trafficStore}->{field_contract_contractID}
		AND A.$self->{trafficStore}->{field_campaignaudio_name} IS NOT NULL
EOQ
	my $result = $self->dbh->prepare($query);
	$result->execute;

	# Clean up path to reflect current operating system.
	my $origin = $self->storage_path."/traffic";
	$origin = File::Spec->canonpath($origin);
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireScheduledMediaCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $old_umask = umask;
	umask 000;

	while (my $line = $result->fetchrow_hashref) {
		my $ext = Retromod::Global->mime2extension($$line{'mimetype'});
		my $mediaID = $$line{'mediaID'};
		my $filename = $$line{'name'};
		my $command = 'rsync -rlvz --exclude ".svn/" "'.$origin.'/'.$mediaID.'" "'.$path.'/'.$filename.'.'.$ext.'"';
		print $command."\n";
		system($command)
	}
	umask $old_umask;
    $result->finish;
    print "Updated local scheduled media cache."."\n";
}

sub cache_schedule {
	my $self = shift;
	$self->purge_cache_schedule();
	$self->update_cache_schedule();
}

# /** Purge local schedule cache. */
sub purge_cache_schedule {
	my $self = shift;

 	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireScheduleCache;
	$path = File::Spec->canonpath($path);

    print "Purging local schedule cache..."."\n";
    my $command = 'rm -f '.Retromod::Global->escape_string($path).'/*';
    print $command."\n";
    system($command);
    print "Purged local schedule cache."."\n";
}

# /** Generate local schedule cache. */
sub update_cache_schedule {
	my $self = shift;

	if (!defined $self->trafficStore) {
		$self->open_remote_storage('traffic');
	}

    print "Updating local schedule cache..."."\n";
	my $query = <<EOQ;
		SELECT A.$self->{trafficStore}->{field_campaignaudio_name} AS name,
			S.$self->{trafficStore}->{field_spot_dateTime} AS dateTime,
			A.$self->{trafficStore}->{field_campaignaudio_mimetype} AS mimetype
		FROM (`$self->{trafficStore}->{table_traffic_spot}` S,
			`$self->{trafficStore}->{table_traffic_campaignaudio}` A,
			`$self->{trafficStore}->{table_traffic_contract}` O
		)
		WHERE (S.$self->{trafficStore}->{field_spot_dateTime} >= TIMESTAMP(NOW() - INTERVAL 2 HOUR)
			&& S.$self->{trafficStore}->{field_spot_dateTime} < TIMESTAMP(NOW() + INTERVAL 4 DAY)
		)
		AND S.$self->{trafficStore}->{field_spot_playedDateTime} IS NULL
		AND (O.$self->{trafficStore}->{field_contract_contractStatus} = 'Approved'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Invoiced'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Paid'
		)
		AND A.$self->{trafficStore}->{field_campaignaudio_productionStatus} = 'Approved'
		AND S.$self->{trafficStore}->{field_spot_inactive} != 'true'
		AND A.$self->{trafficStore}->{field_campaignaudio_filename} != ''
		AND A.$self->{trafficStore}->{field_campaignaudio_mediaID} = S.$self->{trafficStore}->{field_spot_mediaID}
		AND A.$self->{trafficStore}->{field_campaignaudio_contractID} = O.$self->{trafficStore}->{field_contract_contractID}
		AND A.$self->{trafficStore}->{field_campaignaudio_name} IS NOT NULL
EOQ
	my $result = $self->dbh->prepare($query);
	$result->execute;

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireScheduleCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $old_umask = umask;
	umask 000;

	while (my $line = $result->fetchrow_hashref) {
		my $ext = Retromod::Global->mime2extension($$line{'mimetype'});
		my $dateTime = $$line{'dateTime'};
		$dateTime =~ s/(....)-(..)-(..) (..):(..):(..)/\1-\2-\3 \4-\5-\6/g;
		my $command = "touch ".Retromod::Global->escape_string($path)."/".quotemeta($dateTime.";".$$line{'name'}.".".$ext);
		# print $dateTime."\n";
		# print $$line{'name'}.".".$ext."\n";
		# print $command."\n";
		system($command);
	}
	umask $old_umask;
    $result->finish;
    print "Updated local schedule cache."."\n";
}

sub ingest_logs {
	my $self = shift;

	if (!defined $self->trafficStore) {
		$self->open_remote_storage('traffic');
	}

    print "Updating database from local log cache..."."\n";

	my ($playeddate, $date, $updated);

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireLogCache;
	$path = File::Spec->canonpath($path);

	while (defined(my $filename = <$path/*.played>)) {
		print "Reading file: ".$filename."\n";
		my @pieces = $filename =~ /$path\/([^;]*);(.*).played/;
		my $studiofilename = $pieces[0].";".$pieces[1].".played";
		$date = $pieces[0];
		$date =~ s/(....)-(..)-(..) (..)-(..)-(..)/\1-\2-\3 \4:\5:\6/g;
		my $file = $pieces[1];
		open(PLAYEDFILE,$filename) or die("Couldn't open file: ".$filename);
		while (<PLAYEDFILE>) {
			chomp;
			## do something with $_
#            print "File said:", $_, "\n";
			if ($_ != "") {
				$playeddate = $_;
				$playeddate =~ s/(....)-(..)-(..) (..)-(..)-(..)/\1-\2-\3 \4:\5:\6/g;
			}
		}
		close(PLAYEDFILE);

		## Strip the file extension of the filename
		my $name_no_ext = substr $file, 0, -4;

		## now update the database
		$updated = 0;   ## because we don't know which database
		if ($updated == 0) {
			my $query = <<EOQ;
				SELECT S.$self->{trafficStore}->{field_spot_mediaID} AS mediaID,
					S.$self->{trafficStore}->{field_spot_dateTime} AS dateTime
				FROM (`$self->{trafficStore}->{table_traffic_spot}` S,
					`$self->{trafficStore}->{table_traffic_campaignaudio}` A
				)
				WHERE S.$self->{trafficStore}->{field_spot_mediaID} = A.$self->{trafficStore}->{field_campaignaudio_mediaID}
				AND A.$self->{trafficStore}->{field_campaignaudio_name} LIKE '${name_no_ext}%'
				AND S.$self->{trafficStore}->{field_spot_dateTime} = '$date'
EOQ
			my $result = $self->dbh->prepare($query);
			$result->execute;
			if (my $line = $result->fetchrow_hashref) {
				my $queryupdate = <<EOQ;
					UPDATE `$self->{trafficStore}->{table_traffic_spot}`
					SET $self->{trafficStore}->{field_spot_playedDateTime} = '$playeddate'
					WHERE $self->{trafficStore}->{field_spot_mediaID} = '$$line{'mediaID'}'
					AND $self->{trafficStore}->{field_spot_dateTime} = '$$line{'dateTime'}'
EOQ
				my $resultupdate = $self->dbh->prepare($queryupdate);
				$resultupdate->execute;
				$resultupdate->finish;
				$updated = 1;
			}
			$result->finish;
		}

		if ($updated == 0) {
			print "Played file not updated:\n";
			print "    Date: ".$date."\n";
			print "    File: ".$file."\n";
			print "    Played: ".$playeddate."\n";
		} else {
			# Purge local cahce copy.
			my $command = "rm ".$path."/".quotemeta($studiofilename);
			# print $command."\n";
			system($command);
		}
	}
    print "Updated database from local log cache..."."\n";
}

sub sync_remote_logs {
	my $self = shift;

	print "Syncing local logs from remote client '".$self->ip."'.\n";

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireLogCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $command = 'rm -R -f '.$path.'/*';
	print $command."\n";
	system($command);

	my $command = 'rsync -avz --delete --delete-after --remove-source-files '.$self->username.'@'.$self->ip.':"'.$self->log_path.'/" "'.$path.'/"';
	print $command."\n";
	system($command);
	print "Retrieved logs from remote client '".$self->ip."'.\n";
}

sub sync_remote_scheduled_media {
	my $self = shift;

	print "Syncing local scheduled media to remote client '".$self->ip."'.\n";

	#  Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireScheduledMediaCache;
	$path = File::Spec->canonpath($path);

	my $command = 'rsync -avz --delete --delete-after --delete-excluded --ignore-errors '.Retromod::Global->escape_string($path).'/ '.$self->username.'@'.$self->ip.':"'.$self->media_path.'/"';
	print $command."\n";
	system($command);
	print "Copied scheduled media to remote client '".$self->ip."'.\n";
}

sub sync_remote_schedule {
	my $self = shift;

	print "Syncing local schedule to remote client '".$self->ip."'.\n";

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->flywireCache."/".$self->flywireScheduleCache;
	$path = File::Spec->canonpath($path);

	my $command = 'rsync -avz --delete --delete-after --delete-excluded --ignore-errors '.Retromod::Global->escape_string($path).'/ '.$self->username.'@'.$self->ip.':"'.$self->schedule_path.'/"';
	print $command."\n";
	system($command);
	print "Copied schedule to remote client '".$self->ip."'.\n";
}

# /** Checks if a local directory exists, and creates it if missing. */
sub directory_exists {
	my $self = shift;
	my $path = shift;
	# print $path."\n";

	if (-d $path) {
		return 1;
	} else {
		# Does not exist. Try create it instead.
		my $old_umask = umask;
		umask 000;
		print "Creating '".$path."' directory..."."\n";
		mkpath($path, 1, 0777) || return 0;
		umask $old_umask;
		return 1;
	}
}

# /** Test. */
sub test {
	my $self = shift;
	print "This is playout::Client::Flywire->test()!\n";

	# # This is an example of how to use the storage methods from another component
	# #   and the database handler class.
	# my $rh = Retromod->new('traffic', 'share_storage', '');
	# my $t_store = $rh->return();

	# my $dbh = Retromod::DB->connect;
	# my $query = "
		# SELECT ".$t_store->{field_spot_mediaID}." AS mediaID,
			# ".$t_store->{field_spot_dateTime}." AS dateTime
		# FROM (".$t_store->{table_traffic_spot}.")
		# ORDER BY dateTime ASC
		# LIMIT 5
	# ";
	# my $result = $dbh->prepare($query);
	# $result->execute;

	# my %array;
	# print "Example spot schedule: \n";
	# while (my $line = $result->fetchrow_hashref) {
		# my $mediaID = $$line{'mediaID'};
		# my $dateTime = $$line{'dateTime'};

		# print "$mediaID	$dateTime\n";
	# }

}

sub open_remote_storage {
	my $self = shift;
	my $component = shift;

	my $rh = Retromod->new($component, 'share_storage', '');
	my $method = $rh->return();

	if ($method == '0') {
		return 0;
	} else {
		# Load remote storage method for use this session.
		my $property = $component."Store";
		$self->$property($rh->return());
		return 1;
	}
}

sub remote_component_exists {
	my $self = shift;
	my $component = shift;

	my $rh = Retromod->new($component, 'share_storage', '');
	my $method = $rh->return();

	if ($method == '0') {
		return 0;
	} else {
		return 1;
	}
}

sub check_component_exists {
	my $self = shift;
	my $component = shift;

	if ($self->remote_component_exists($component)) {
		return 1;
	} else {
		die playout::Exception->new("Cannot access the prerequisite component '$component'. Cannot proceed. Please install it before executing this command again.");
	}
}

################## TEST #####################

# /** Update production processes on remote Flywire client. */
sub process_all_production {
	my $self = shift;

	$self->process_production_media();
}

# /** Update production media for remote Flywire client. */
sub process_production_media {
	my $self = shift;

	$self->check_component_exists('production');

	$self->update_cache_production_media();
	$self->update_cache_promo_media();
	$self->sync_remote_production_media();
}

sub update_cache_production_media {
	my $self = shift;

	if (!defined $self->productionStore) {
		$self->open_remote_storage('production');
	}

	# Extract production media from data storage.
    print "Updating local production media cache..."."\n";
	my $query = <<EOQ;
		SELECT DISTINCT M.$self->{productionStore}->{production_media_name} AS name,
			M.$self->{productionStore}->{production_media_mediaID} AS mediaID,
			M.$self->{productionStore}->{production_media_mimetype} AS mimetype,
			C.$self->{productionStore}->{production_category_name} AS categoryName
		FROM (`$self->{productionStore}->{production_media}` M,
			`$self->{productionStore}->{production_category}` C,
			`$self->{productionStore}->{production_mediacategory}` A
		)
		WHERE M.$self->{productionStore}->{production_media_productionID} = A.$self->{productionStore}->{production_mediacategory_productionID}
		AND C.$self->{productionStore}->{production_category_categoryID} = A.$self->{productionStore}->{production_mediacategory_categoryID}
		AND M.$self->{productionStore}->{production_media_start} <= CURRENT_DATE
		AND (M.$self->{productionStore}->{production_media_end} IS NULL
			|| M.$self->{productionStore}->{production_media_end} >= CURRENT_DATE
		)
		AND M.$self->{productionStore}->{production_media_filename} != ''
		AND M.$self->{productionStore}->{production_media_name} IS NOT NULL
EOQ
	my $result = $self->dbh->prepare($query);
	$result->execute;

	# Clean up paths to reflect current operating system.
	my $origin = $self->storage_path."/production";
	$origin = File::Spec->canonpath($origin);
	my $path = $self->cache_path."/production/".$self->flywireCache."/".$self->flywireStationMediaCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $old_umask = umask;
	umask 000;

	while (my $line = $result->fetchrow_hashref) {
		my $ext = Retromod::Global->mime2extension($$line{'mimetype'});
		my $mediaID = $$line{'mediaID'};
		my $filename = $$line{'name'};
		my $command = 'rsync -rlvz --exclude ".svn/" "'.$origin.'/'.$mediaID.'" "'.$path.'/'.$filename.'.'.$ext.'"';
		print $command."\n";
		system($command)
	}
	umask $old_umask;
    $result->finish;
    print "Updated local production media cache."."\n";
}

sub update_cache_promo_media {
	my $self = shift;

	if (!defined $self->productionStore) {
		$self->open_remote_storage('production');
	}

	# Extract production promo from data storage.
    print "Updating local production promo cache..."."\n";
	my $query = <<EOQ;
		SELECT DISTINCT M.$self->{productionStore}->{production_media_name} AS name,
			M.$self->{productionStore}->{production_media_mediaID} AS mediaID,
			M.$self->{productionStore}->{production_media_mimetype} AS mimetype
		FROM (`$self->{productionStore}->{production_promo}` O,
			`$self->{productionStore}->{production_media}` M
		)
		WHERE O.$self->{productionStore}->{production_promo_productionID} = M.$self->{productionStore}->{production_media_productionID}
		AND M.$self->{productionStore}->{production_media_start} <= CURRENT_DATE
		AND (M.$self->{productionStore}->{production_media_end} IS NULL
			|| M.$self->{productionStore}->{production_media_end} >= CURRENT_DATE
		)
		AND O.$self->{productionStore}->{production_promo_authorisedStatus} = 'Approved'
		AND M.$self->{productionStore}->{production_media_filename} != ''
		AND M.$self->{productionStore}->{production_media_name} IS NOT NULL
EOQ
	my $result = $self->dbh->prepare($query);
	$result->execute;

	# Clean up paths to reflect current operating system.
	my $origin = $self->storage_path."/production";
	$origin = File::Spec->canonpath($origin);
	my $path = $self->cache_path."/production/".$self->flywireCache."/".$self->flywireStationMediaCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $old_umask = umask;
	umask 000;

	while (my $line = $result->fetchrow_hashref) {
		my $ext = Retromod::Global->mime2extension($$line{'mimetype'});
		my $mediaID = $$line{'mediaID'};
		my $filename = $$line{'name'};
		my $command = 'rsync -rlvz --exclude ".svn/" "'.$origin.'/PROMO-'.$mediaID.'" "'.$path.'/'.$filename.'.'.$ext.'"';
		print $command."\n";
		system($command)
	}
	umask $old_umask;
    $result->finish;
    print "Updated local production promo cache."."\n";
}

sub sync_remote_production_media {
	my $self = shift;

	print "Syncing local production media to remote client '".$self->ip."'.\n";

	#  Clean up path to reflect current operating system.
	my $path = $self->cache_path."/production/".$self->flywireCache."/".$self->flywireStationMediaCache;
	$path = File::Spec->canonpath($path);

	my $command = 'rsync -avz --delete --delete-after --delete-excluded --ignore-errors '.Retromod::Global->escape_string($path).'/ '.$self->username.'@'.$self->ip.':"'.$self->media_path.'/"';
	print $command."\n";
	system($command);
	print "Copied scheduled media to remote client '".$self->ip."'.\n";
}

1;