# $Id$

# /** @file Bentokit Retromod scheduled media and spot update to Novation playout client. */

use playout::Client;

# /** Update module for the scheduled sponsorship data that goes to and from Novation on the client machine. */
package playout::Client::Novation;

# Specify parent for inheritance.
@ISA = (playout::Client);

use strict;
use playout::Storage;
use File::Path;

our $AUTOLOAD;

## Permitted properties.
my %fields = (
	ip => undef,
	log_path => undef,
	media_path => undef,
	schedule_path => undef,
	username => undef,
	onair => undef,
	storage_path => undef,
	cache_path => undef,
	storage_method => undef,
	dbh => undef,
	store => undef,
	trafficStore => undef,
	novationCache => "novation",
	novationAutoMediaCache => "media",
	novationAutoMediaMusicCache => "media",
	novationLogCache => "log",
	novationScheduleCache => "schedule",
	novationScheduledMediaCache => "media",
	novationShowMediaCache => "media",
	novationStationMediaCache => "media",
);

# /** Destructor. */
sub DESTROY {
	#Retromod::DB->disconnect($self->dbh);
	my ($self) = @_;
}

# /** Constructor. */
sub new {
	# Use our inherited AUTOLOAD but allow set and get of our params.
	my $class = shift;
	my $self = $class->SUPER::new();
	my($element);
	foreach $element (keys %fields) {
		$self->{_permitted}->{$element} = $fields{$element};
	}
	@{$self}{keys %fields} = values %fields;

	# Load recurring objects.
	#$self->store = playout::Storage->new();
	#$self->trafficStore = traffic::Storage->new();
	#$self->dbh = Retromod::DB->connect;
	return $self;
}

# /** Help page. */
sub help {
	my($self, $args) = @_;
	# /** @todo If client argument exists print client's help method instead. */
	my $help_message = <<EOH;
This is the help for the Novation client module for the playout component of Retromod.

Usage: update.pl playout method Novation [ arguments ]

  Method:
  help
    Returns this help page.
  process_all
    Updates traffic, ... on the remote Novation client.
  process_all_traffic
    Updates logs, schedule and scheduled media on the remote Novation client.
  process_logs
    Updates logs on the remote Novation client.
  process_scheduled_media
    Updates scheduled media on the remote Novation client.
  process_schedule
    Updates the schedule on the remote Novation client.

EOH
    print "$help_message\n";
}

# /** Update remote Novation client. */
sub process_all {
	my $self = shift;

	$self->process_logs();
	$self->process_scheduled_media();
	$self->process_schedule();
}

# /** Update traffic processes on remote Novation client. */
sub process_all_traffic {
	my $self = shift;

	$self->process_logs();
	$self->process_scheduled_media();
	$self->process_schedule();
}

# /** Update logs for remote Novation client. */
sub process_logs {
	my $self = shift;

	# Ingest logs only if studio was designated onair.
	if ($self->onair) {
		print "Remote client '".$self->ip."' was designated as on-air. Proceeding...\n";
		$self->sync_remote_logs();
		$self->ingest_logs();
	} else {
		print "Remote client '".$self->ip."' was not designated as on-air. Logs not ingested.\n";
	}
}

# /** Update scheduled media for remote Novation client. */
sub process_scheduled_media {
	my $self = shift;
	$self->cache_scheduled_media();
	$self->sync_remote_scheduled_media();
}

# /** Update schedule for remote Novation client. */
sub process_schedule {
	my $self = shift;
	$self->cache_schedule();
	$self->sync_remote_schedule();
}

# /** Update local cache for scheduled media. */
sub cache_scheduled_media {
	my $self = shift;
	$self->purge_cache_scheduled_media();
	$self->update_cache_scheduled_media();
}

sub purge_cache_scheduled_media {
	my $self = shift;

	# Purge previously cached media on the server.

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationScheduledMediaCache;
	$path = File::Spec->canonpath($path);

    print "Purging local scheduled media cache..."."\n";
    my $command = 'rm -R -f '.Retromod::Global->escape_string($path).'/*';
    print $command."\n";
    system($command);
    print "Purged local scheduled media cache."."\n";
}

sub update_cache_scheduled_media {
	my $self = shift;

	# Extract scheduled media from data storage.
    print "Updating local scheduled media cache..."."\n";
	my $query = <<EOQ;
		SELECT DISTINCT A.$self->{trafficStore}->{field_campaignaudio_mediaID} AS mediaID,
			A.$self->{trafficStore}->{field_campaignaudio_campaignAudioID} AS campaignAudioID,
			A.$self->{trafficStore}->{field_campaignaudio_mimetype} AS mimetype
		FROM (`$self->{trafficStore}->{table_traffic_spot}` S,
			`$self->{trafficStore}->{table_traffic_campaignaudio}` A,
			`$self->{trafficStore}->{table_traffic_contract}` O
		)
		WHERE (S.$self->{trafficStore}->{field_spot_dateTime} >= TIMESTAMP(NOW() - INTERVAL 2 HOUR)
			&& S.$self->{trafficStore}->{field_spot_dateTime} < TIMESTAMP(NOW() + INTERVAL 4 DAY)
		)
		AND (O.$self->{trafficStore}->{field_contract_contractStatus} = 'Approved'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Invoiced'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Paid'
		)
		AND A.$self->{trafficStore}->{field_campaignaudio_productionStatus} = 'Approved'
		AND S.$self->{trafficStore}->{field_spot_inactive} != 'true'
		AND A.$self->{trafficStore}->{field_campaignaudio_filename} != ''
		AND (A.$self->{trafficStore}->{field_campaignaudio_mimetype} = 'audio/x-wav'
			|| A.$self->{trafficStore}->{field_campaignaudio_mimetype} = 'audio/wav'
		)
		AND A.$self->{trafficStore}->{field_campaignaudio_mediaID} = S.$self->{trafficStore}->{field_spot_mediaID}
		AND A.$self->{trafficStore}->{field_campaignaudio_contractID} = O.$self->{trafficStore}->{field_contract_contractID}
EOQ
	my $result = $self->dbh->prepare($query);
	$result->execute;

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationScheduledMediaCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $old_umask = umask;
	umask 000;

	while (my $line = $result->fetchrow_hashref) {
		my $ext = Retromod::Global->mime2extension($$line{'mimetype'});
		my $mediaID = $$line{'mediaID'};
		my $campaignAudioID = $$line{'campaignAudioID'};
		my $command = 'rsync -rltvz --exclude ".svn/" "'.$self->storage_path.'/traffic/'.$mediaID.'" "'.$path.'/'.$campaignAudioID.'.'.$ext.'"';
		print $command."\n";
		system($command)
	}
	umask $old_umask;
    $result->finish;
    print "Updated local scheduled media cache."."\n";
}

sub cache_schedule {
	my $self = shift;
	$self->purge_cache_schedule();
	$self->update_cache_schedule();
}

# /** Purge local schedule cache. */
sub purge_cache_schedule {
	my $self = shift;

 	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationScheduleCache;
	$path = File::Spec->canonpath($path);

    print "Purging local schedule cache..."."\n";
    my $command = 'rm -f '.Retromod::Global->escape_string($path).'/*';
    print $command."\n";
    system($command);
    print "Purged local schedule cache."."\n";
}

# /** Generate local schedule cache. */
sub update_cache_schedule {
	my $self = shift;

    print "Updating local schedule cache..."."\n";
	my $query = <<EOQ;
		SELECT M.$self->{trafficStore}->{field_campaign_name} AS name,
			S.$self->{trafficStore}->{field_spot_dateTime} AS dateTime,
			A.$self->{trafficStore}->{field_campaignaudio_campaignAudioID} AS campaignAudioID,
			A.$self->{trafficStore}->{field_campaignaudio_mimetype} AS mimetype,
			S.$self->{trafficStore}->{field_spot_priority} AS priority
		FROM (`$self->{trafficStore}->{table_traffic_spot}` S,
			`$self->{trafficStore}->{table_traffic_campaignaudio}` A,
			`$self->{trafficStore}->{table_traffic_contract}` O,
			`$self->{trafficStore}->{table_traffic_campaign}` M
		)
		WHERE (S.$self->{trafficStore}->{field_spot_dateTime} >= TIMESTAMP(NOW() - INTERVAL 168 HOUR)
			&& S.$self->{trafficStore}->{field_spot_dateTime} < TIMESTAMP(NOW() + INTERVAL 4 DAY)
		)
		AND (O.$self->{trafficStore}->{field_contract_contractStatus} = 'Approved'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Invoiced'
			|| O.$self->{trafficStore}->{field_contract_contractStatus} = 'Paid'
		)
		AND A.$self->{trafficStore}->{field_campaignaudio_productionStatus} = 'Approved'
		AND S.$self->{trafficStore}->{field_spot_inactive} != 'true'
		AND A.$self->{trafficStore}->{field_campaignaudio_filename} != ''
		AND (A.$self->{trafficStore}->{field_campaignaudio_mimetype} = 'audio/x-wav'
			|| A.$self->{trafficStore}->{field_campaignaudio_mimetype} = 'audio/wav'
		)
		AND A.$self->{trafficStore}->{field_campaignaudio_mediaID} = S.$self->{trafficStore}->{field_spot_mediaID}
		AND A.$self->{trafficStore}->{field_campaignaudio_contractID} = O.$self->{trafficStore}->{field_contract_contractID}
		AND A.$self->{trafficStore}->{field_campaignaudio_campaignID} = M.$self->{trafficStore}->{field_campaign_campaignID}
		AND A.$self->{trafficStore}->{field_campaignaudio_name} IS NOT NULL
		ORDER BY dateTime, priority
EOQ
	my $result = $self->dbh->prepare($query);
	$result->execute;

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationScheduleCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $old_umask = umask;
	umask 000;

	# Create an array with a key for every sequence hour in a month (i.e. 31 * 24).
	my @all_seq = (1 .. 744);

	# Blank all sequence files.
	my $seqnumber;
	foreach $seqnumber (@all_seq) {
		# /* Append a blank footer to the end of the file to make Novation display empty carts */
		open(SEQFILE,'>'.$path."/SEQ".$seqnumber.".SEQ");
		close(SEQFILE);
	}

	# Create array of modified SEQ files so we can append a footer on them later.
	my %touched = ();
	my ($dateTime, $year, $mon, $day, $hour, $min, $sec, $seq, $seqfile);
	my (@dash, @space, @col);
	while (my $line = $result->fetchrow_hashref) {
		$dateTime = $$line{'dateTime'};

		# /* Split spot schedule time into usable components */
		@dash = split(/-/, $dateTime);
		$year = $dash[0];
		$year = $year - 1900;
		$mon = $dash[1];
		$mon = $mon - 1;
		@space = split(/ /,$dash[2]);
		$day = $space[0];
		@col = split(/:/, $space[1]);
		$hour = $col[0];
		$hour = $hour - 1;
		$min = $col[1];
		$sec = $col[2];

		# /* Calculate the Novation sequence file number */
		my $seq = ($day * 24) - 24 + 1 + ($hour + 1);
		$touched{$seq} = 1;

		# /* Create/open and write/append the spot to the sequence file */
		open(SEQFILE,'>>'.$path."/SEQ".$seq.".SEQ");
		# print("Added spot to: SEQ".$seq.".SEQ"."\n");
		# TODO: change this client name ?!
		print SEQFILE $$line{'campaignAudioID'}." ,".$$line{'name'}."\r\n";
		print SEQFILE ' 8'."\r\n";

		close(SEQFILE);
	}
    $result->finish;
	my $footer_count;

	# Retrieve unique sequence files modified.
	my @seq_files = keys(%touched);

	foreach $seqfile (@seq_files) {
		# /* Append a blank footer to the end of the file to make Novation display empty carts */
		open(SEQFILE,'>>'.$path."/SEQ".$seqfile.".SEQ");
		for ($footer_count=0; $footer_count<10; $footer_count++) {
			print SEQFILE '?'."\r\n";
			print SEQFILE ' 8'."\r\n";
		}
		close(SEQFILE);
		# print("Appended footer to: SEQ".$seqfile.".SEQ"."\n");
		print("Modified schedule file: SEQ".$seqfile.".SEQ"."\n");
	}

	# Novation requires a sequence file for every hour, regardless of if there is scheduled files to play.
	# Therefore we need to create blank sequence files for every hour.

	# Determine which sequence hours have not been scheduled to, by diff checking the touched array.
	my ($e, $newseq);
	my %count;
	my @diff;

	foreach $e (@all_seq, @seq_files) {
		$count{$e}++
	}
	foreach $e (keys %count) {
		if ($count{$e} != 2) {
			push @diff, $e;
		}
	}

	foreach $newseq (@diff) {
		# /* Append a blank footer to the end of the file to make Novation display empty carts */
		open(SEQFILE,'>>'.$path."/SEQ".$newseq.".SEQ");
		print SEQFILE '0000   ,NO SPOTS THIS HOUR'."\r\n";
		print SEQFILE ' 8'."\r\n";
		for ($footer_count=0; $footer_count<10; $footer_count++) {
			print SEQFILE '?'."\r\n";
			print SEQFILE ' 8'."\r\n";
		}
		close(SEQFILE);
		# print("Blanking schedule file: SEQ".$newseq.".SEQ"."\n");
	}

	umask $old_umask;
    print "Updated local schedule cache."."\n";
}

sub ingest_logs {
	my $self = shift;

    print "Updating database from local log cache..."."\n";

	my ($playeddatetime, $date, $updated, $campaignAudioID);
	my @playout_log;

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationLogCache;
	$path = File::Spec->canonpath($path);

	my $logfile = $path."/Played.txt";
	$logfile = File::Spec->canonpath($logfile);

	my $success = 1;

	open (LOGFILE, $logfile) or $success = 0;

	if ($success) {

		while (<LOGFILE>) {
			chomp;
			## do something with $_
			# print "File said :", $_, "\n";
			if ($_ != "") {
				## Split each line by whitespace
				my @line = split(/\s+/, $_);

				my $period = $line[2];
				my $unknown = $line[3];
				$campaignAudioID = $line[4];

				## Transform date.
				$line[0] =~ s/(..)\/(..)\/(....)/\3-\2-\1/g;
				$date = $line[0];
				# print "mate ".$date."\n";

				## Convert to 24 hour time.
				my @t = split(/:/, $line[1]);
				if ($period == 'PM' && $t[0] < 12) {
					$t[0] = $t[0] + 12;
				}
				if ($period == 'AM' && $t[0] == 12) {
					$t[0] = 0;
				}
				my $military_time = sprintf "%02d:%02d:%02d", $t[0], $t[1], $t[2];
				# print "military_time ".$military_time."\n";

				## Assemble played datetime.
				$playeddatetime = $date." ".$military_time;
			}

			## Update the database.
			$updated = 0;   ## because we don't know which database
			if ($updated == 0) {
				## Because we don't have the scheduled spot time, we need to make a guess as to the correct spot that was played.
				## Find the spot closest to the played time.
				my $query = <<EOQ;
					SELECT S.$self->{trafficStore}->{field_spot_dateTime} AS dateTime,
						S.$self->{trafficStore}->{field_spot_mediaID} AS mediaID,
						ABS(TIME_TO_SEC(TIMEDIFF(S.$self->{trafficStore}->{field_spot_dateTime}, '${playeddatetime}'))) AS rank
					FROM (`$self->{trafficStore}->{table_traffic_spot}` S,
						`$self->{trafficStore}->{table_traffic_campaignaudio}` A
					)
					WHERE S.$self->{trafficStore}->{field_spot_mediaID} = A.$self->{trafficStore}->{field_campaignaudio_mediaID}
					AND A.$self->{trafficStore}->{field_campaignaudio_campaignAudioID} = '$campaignAudioID'
					AND (S.$self->{trafficStore}->{field_spot_playedDateTime} = ''
						|| S.$self->{trafficStore}->{field_spot_playedDateTime} IS NULL
					)
					ORDER BY rank
					LIMIT 1
EOQ
				# print $query;
				my $result = $self->dbh->prepare($query);
				$result->execute;
				if (my $line = $result->fetchrow_hashref) {
					# print "mediaID ".$$line{'mediaID'}." dateTime ".$$line{'dateTime'}." rank ".$$line{'rank'}."\n";
					my $queryupdate = <<EOQ;
						UPDATE `$self->{trafficStore}->{table_traffic_spot}`
						SET $self->{trafficStore}->{field_spot_playedDateTime} = '$playeddatetime'
						WHERE $self->{trafficStore}->{field_spot_mediaID} = '$$line{'mediaID'}'
						AND $self->{trafficStore}->{field_spot_dateTime} = '$$line{'dateTime'}'
EOQ
					# print $queryupdate;
					my $resultupdate = $self->dbh->prepare($queryupdate);
					$resultupdate->execute;
					$resultupdate->finish;
					$updated = 1;
				} else {
					print "No results\n";
				}
				$result->finish;
			}

			if ($updated == 0) {
				print "Logs not updated for:\n";
				print "    campaignAudioID: ".$campaignAudioID."\n";
				print "    Date: ".$playeddatetime."\n";
			}
		}
		close(LOGFILE);

		print "Database updated."."\n";

		# Purge local cache copy.
		my $command = "rm ".$logfile;
		print $command."\n";
		system($command);
		print "Updated database from local log cache..."."\n";
	} else {
		print "Could not open Played.txt. Database not updated."."\n";
	}

}

sub sync_remote_logs {
	my $self = shift;

	print "Syncing local logs from remote client '".$self->ip."'.\n";

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationLogCache;
	$path = File::Spec->canonpath($path);

	if (!$self->directory_exists($path)) {
		print "Could not create '".$path."' directory. Process failed."."\n";
		exit();
	}

	my $command = 'rm -R -f '.$path.'/*';
	print $command."\n";
	system($command);

	my $command = 'rsync -avz --delete --delete-after --remove-source-files '.$self->username.'@'.$self->ip.':"'.$self->log_path.'/Played.txt" "'.$path.'/"';
	print $command."\n";
	system($command);
	print "Retrieved logs from remote client '".$self->ip."'.\n";
}

sub sync_remote_scheduled_media {
	my $self = shift;

	print "Syncing local scheduled media to remote client '".$self->ip."'.\n";

	#  Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationScheduledMediaCache;
	$path = File::Spec->canonpath($path);

	my $command = 'rsync -avz '.Retromod::Global->escape_string($path).'/ '.$self->username.'@'.$self->ip.':"'.$self->media_path.'/"';
	print $command."\n";
	system($command);
	print "Copied scheduled media to remote client '".$self->ip."'.\n";
}

sub sync_remote_schedule {
	my $self = shift;

	print "Syncing local schedule to remote client '".$self->ip."'.\n";

	# Clean up path to reflect current operating system.
	my $path = $self->cache_path."/traffic/".$self->novationCache."/".$self->novationScheduleCache;
	$path = File::Spec->canonpath($path);

	my $command = 'rsync -avz --delete --delete-after --delete-excluded --ignore-errors '.Retromod::Global->escape_string($path).'/ '.$self->username.'@'.$self->ip.':"'.$self->schedule_path.'/"';
	print $command."\n";
	system($command);
	print "Copied schedule to remote client '".$self->ip."'.\n";
}

# /** Checks if a local directory exists, and creates it if missing. */
sub directory_exists {
	my $self = shift;
	my $path = shift;
	# print $path."\n";

	if (-d $path) {
		return 1;
	} else {
		# Does not exist. Try create it instead.
		my $old_umask = umask;
		umask 000;
		print "Creating '".$path."' directory..."."\n";
		mkpath($path, 1, 0777) || return 0;
		umask $old_umask;
		return 1;
	}
}


# /** Test. */
sub update_logs {
	my $self = shift;
	print "This is playout::Client::Novation->update_logs()!\n";

	# This is an example of how to use the storage methods from another component
	#   and the database handler class.
	my $rh = Retromod->new('traffic', 'share_storage', '');
	my $t_store = $rh->return();

	my $dbh = Retromod::DB->connect;
	my $query = "
		SELECT ".$t_store->{field_spot_mediaID}." AS mediaID,
			".$t_store->{field_spot_dateTime}." AS dateTime
		FROM (".$t_store->{table_traffic_spot}.")
		ORDER BY dateTime ASC
		LIMIT 5
	";
	my $result = $dbh->prepare($query);
	$result->execute;

	my %array;
	print "Example spot schedule: \n";
	while (my $line = $result->fetchrow_hashref) {
		my $mediaID = $$line{'mediaID'};
		my $dateTime = $$line{'dateTime'};

		print "$mediaID	$dateTime\n";
	}

}

# /** Update local caches. */
sub update_cache {
	my $self = shift;
#	my $that = shift;
#	my $self = ref($that);
	#$self = playout::Client::Novation->new();

	#$self->update_cache_schedule();
	#$self->update_cache_scheduled_media();
	return $self->test_majora();
	#return;
}

sub test_majora {
	my $self = shift;
	print "HELLO WORLD!\n";
	# print $self->purge_cache_schedule();
	#return "HELLO WORLD!";

	#throw playoutException("Throwing errors.");
}

1;